@@ -22,17 +22,20 @@ module System.Console.Terminfo.Base(
                             tiGetNum,
                             tiGetStr,
                             -- * Output
+                            -- $outputdoc
+                            tiGetOutput1,
+                            OutputCap,
+                            TermStr,
+                            -- ** TermOutput
                             TermOutput(),
                             runTermOutput,
                             hRunTermOutput,
                             termText,
                             tiGetOutput,
                             LinesAffected,
-                            tiGetOutput1,
-                            OutputCap,
                             -- ** Monoid functions
                             Monoid(..),
-                            (<#>)
+                            (<#>),
                             ) where
 
 
@@ -50,7 +53,7 @@ import Control.Exception.Extensible
 import Data.Typeable
 
 
-data TERMINAL = TERMINAL
+data TERMINAL
 newtype Terminal = Terminal (ForeignPtr TERMINAL)
 
 foreign import ccall "&" cur_term :: Ptr (Ptr TERMINAL)
@@ -107,49 +110,107 @@ setupTermFromEnv = do
     handleBadEnv :: IOException -> IO String
     handleBadEnv _ = return ""
 
+-- TODO: this isn't really thread-safe...
 withCurTerm :: Terminal -> IO a -> IO a
 withCurTerm (Terminal term) f = withForeignPtr term $ \cterm -> do
         old_term <- peek cur_term
         if old_term /= cterm
             then do
-                    set_curterm cterm
+                    _ <- set_curterm cterm
                     x <- f
-                    set_curterm old_term
+                    _ <- set_curterm old_term
                     return x
             else f
 
+
+----------------------
+
+-- Note I'm relying on this working even for strings with unset parameters.
+strHasPadding :: String -> Bool
+strHasPadding [] = False
+strHasPadding ('$':'<':_) = True
+strHasPadding (_:cs) = strHasPadding cs
+
+-- | An action which sends output to the terminal.  That output may mix plain text with control
+-- characters and escape sequences, along with delays (called \"padding\") required by some older
+-- terminals.
+
+-- We structure this similarly to ShowS, so that appends don't cause space leaks.
+newtype TermOutput = TermOutput ([TermOutputType] -> [TermOutputType])
+
+data TermOutputType = TOCmd LinesAffected String
+                    | TOStr String
+
+instance Monoid TermOutput where
+    mempty = TermOutput id
+    TermOutput xs `mappend` TermOutput ys = TermOutput (xs . ys)
+
+termText :: String -> TermOutput 
+termText str = TermOutput (TOStr str :)
+
+-- | Write the terminal output to the standard output device.
+runTermOutput :: Terminal -> TermOutput -> IO ()
+runTermOutput = hRunTermOutput stdout
+
+-- | Write the terminal output to the terminal or file managed by the given
+-- 'Handle'.
+hRunTermOutput :: Handle -> Terminal -> TermOutput -> IO ()
+hRunTermOutput h term (TermOutput to) = do
+    putc_ptr <- mkCallback putc
+    withCurTerm term $ mapM_ (writeToTerm putc_ptr h) (to [])
+    freeHaskellFunPtr putc_ptr
+    hFlush h
+  where
+    putc c = let c' = toEnum $ fromEnum c
+             in hPutChar h c' >> hFlush h >> return c
+
+-- NOTE: Currently, the output is checked every time tparm is called.
+-- It might be faster to check for padding once in tiGetOutput1.
+writeToTerm :: FunPtr CharOutput -> Handle -> TermOutputType -> IO ()
+writeToTerm putc h (TOCmd numLines s)
+    | strHasPadding s = tPuts s numLines putc
+    | otherwise = hPutStr h s
+writeToTerm _ h (TOStr s) = hPutStr h s
+
+infixl 2 <#>
+
+-- | An operator version of 'mappend'.
+(<#>) :: Monoid m => m -> m -> m
+(<#>) = mappend
+---------------------------------
+
 -- | A feature or operation which a 'Terminal' may define.
-newtype Capability a = Capability (IO (Maybe a))
+newtype Capability a = Capability (Terminal -> IO (Maybe a))
 
 getCapability :: Terminal -> Capability a -> Maybe a
-getCapability term (Capability f) = unsafePerformIO $ withCurTerm term f
+getCapability term (Capability f) = unsafePerformIO $ withCurTerm term (f term)
 
 -- Note that the instances for Capability of Functor, Monad and MonadPlus 
 -- use the corresponding instances for Maybe.
 instance Functor Capability where
-    fmap f (Capability g) = Capability (fmap (fmap f) g) 
+    fmap f (Capability g) = Capability $ \t -> fmap (fmap f) (g t)
 
 instance Monad Capability where
-    return = Capability . return . Just
-    Capability f >>= g = Capability $ do
-        mx <- f
+    return = Capability . const . return . Just
+    Capability f >>= g = Capability $ \t -> do
+        mx <- f t
         case mx of
             Nothing -> return Nothing
-            Just x -> let Capability g' = g x in g'
+            Just x -> let Capability g' = g x in g' t
 
 instance MonadPlus Capability where
-    mzero = Capability (return Nothing)
-    Capability f `mplus` Capability g = Capability $ do
-        mx <- f
+    mzero = Capability (const $ return Nothing)
+    Capability f `mplus` Capability g = Capability $ \t -> do
+        mx <- f t
         case mx of
-            Nothing -> g
+            Nothing -> g t
             _ -> return mx
 
 foreign import ccall tigetnum :: CString -> IO CInt
 
 -- | Look up a numeric capability in the terminfo database.
 tiGetNum :: String -> Capability Int 
-tiGetNum cap = Capability $ do
+tiGetNum cap = Capability $ const $ do
                 n <- fmap fromEnum (withCString cap tigetnum)
                 if n >= 0
                     then return (Just n)
@@ -162,7 +223,7 @@ foreign import ccall tigetflag :: CString -> IO CInt
 -- capability is absent or set to false, and returns 'True' otherwise.  
 -- 
 tiGetFlag :: String -> Capability Bool
-tiGetFlag cap = Capability $ fmap (Just . (>0)) $
+tiGetFlag cap = Capability $ const $ fmap (Just . (>0)) $
                         withCString cap tigetflag
                 
 -- | Look up a boolean capability in the terminfo database, and fail if
@@ -172,12 +233,11 @@ tiGuardFlag cap = tiGetFlag cap >>= guard
                 
 foreign import ccall tigetstr :: CString -> IO CString
 
--- | Look up a string capability in the terminfo database.  
---
--- Note: Do not use this function for terminal output; use 'tiGetOutput'
--- instead.
+{-# DEPRECATED tiGetStr "use tiGetOutput instead." #-} 
+-- | Look up a string capability in the terminfo database.  NOTE: This function is deprecated; use
+-- 'tiGetOutput1' instead.
 tiGetStr :: String -> Capability String
-tiGetStr cap = Capability $ do
+tiGetStr cap = Capability $ const $ do
                 result <- withCString cap tigetstr 
                 if result == nullPtr || result == neg1Ptr
                     then return Nothing
@@ -185,30 +245,42 @@ tiGetStr cap = Capability $ do
     where
         -- hack; tigetstr sometimes returns (-1)
         neg1Ptr = nullPtr `plusPtr` (-1)
+
+
+---------------
+
+
                     
 foreign import ccall tparm ::
     CString -> CLong -> CLong -> CLong -> CLong -> CLong -> CLong 
     -> CLong -> CLong -> CLong -- p1,...,p9
     -> IO CString
 
-
 -- Note: I may want to cut out the middleman and pipe tGoto/tGetStr together
 -- with tput without a String marshall in the middle.
 -- directly without 
 
-tParm :: String -> [Int] -> IO String
-tParm cap ps = tparm' (map toEnum ps ++ repeat 0)
+tParm :: String -> Capability ([Int] -> String)
+tParm cap = Capability $ \t -> return $ Just 
+        $ \ps -> unsafePerformIO $ withCurTerm t $
+                    tparm' (map toEnum ps ++ repeat 0)
     where tparm' (p1:p2:p3:p4:p5:p6:p7:p8:p9:_)
             = withCString cap $ \c_cap -> do
                 result <- tparm c_cap p1 p2 p3 p4 p5 p6 p7 p8 p9
-                peekCString result
+                if result == nullPtr
+                    then return ""
+                    else peekCString result
+          tparm' _ = fail "tParm: List too short"
 
 -- | Look up an output capability in the terminfo database.  
 tiGetOutput :: String -> Capability ([Int] -> LinesAffected -> TermOutput)
-tiGetOutput cap = flip fmap (tiGetStr cap) $ 
-    \str ps la -> TermOutput $ \_ putc -> do
-        outStr <- tParm str ps
-        tPuts outStr la putc
+tiGetOutput cap = do
+    str <- tiGetStr cap
+    f <- tParm str
+    return $ \ps la -> fromStr la $ f ps
+
+fromStr :: LinesAffected -> String -> TermOutput
+fromStr la s = TermOutput (TOCmd la s :)
 
 type CharOutput = CInt -> IO CInt
 
@@ -226,56 +298,64 @@ type LinesAffected = Int
 tPuts :: String -> LinesAffected -> FunPtr CharOutput -> IO ()
 tPuts s n putc = withCString s $ \c_str -> tputs c_str (toEnum n) putc
 
--- | An action which sends output to the terminal.  That output may mix plain text with control
--- characters and escape sequences, along with delays (called \"padding\") required by some older
--- terminals.
-newtype TermOutput = TermOutput (Handle -> FunPtr CharOutput -> IO ())
-
--- | Write the terminal output to the standard output device.
-runTermOutput :: Terminal -> TermOutput -> IO ()
-runTermOutput = hRunTermOutput stdout
-
--- | Write the terminal output to the terminal or file managed by the given
--- 'Handle'.
-hRunTermOutput :: Handle -> Terminal -> TermOutput -> IO ()
-hRunTermOutput h term (TermOutput to) = do
-    putc_ptr <- mkCallback putc
-    withCurTerm term (to h putc_ptr)
-    freeHaskellFunPtr putc_ptr
-  where
-    putc c = let c' = toEnum $ fromEnum c
-             in hPutChar h c' >> hFlush h >> return c
-
--- | Output plain text containing no control characters or escape sequences.
-termText :: String -> TermOutput
-termText str = TermOutput $ \h _ -> hPutStr h str >> hFlush h
-
-instance Monoid TermOutput where 
-    mempty = TermOutput $ \_ _ -> return ()
-    TermOutput f `mappend` TermOutput g = TermOutput $ \h putc -> f h putc >> g h putc 
-
--- | A type class to encapsulate capabilities which take in zero or more 
--- parameters.
-class OutputCap f where
-    outputCap :: ([Int] -> TermOutput) -> [Int] -> f
-
-instance OutputCap TermOutput where
-    outputCap f xs = f (reverse xs)
-
-instance (Enum a, OutputCap f) => OutputCap (a -> f) where
-    outputCap f xs = \x -> outputCap f (fromEnum x:xs)
 
 -- | Look up an output capability which takes a fixed number of parameters
 -- (for example, @Int -> Int -> TermOutput@).
 -- 
 -- For capabilities which may contain variable-length
 -- padding, use 'tiGetOutput' instead.
-tiGetOutput1 :: OutputCap f => String -> Capability f
-tiGetOutput1 str = fmap (\f -> outputCap (flip f 1) []) $ tiGetOutput str
+tiGetOutput1 :: forall f . OutputCap f => String -> Capability f
+tiGetOutput1 str = do
+    cap <- tiGetStr str
+    guard (hasOkPadding (undefined :: f) cap)
+    f <- tParm cap
+    return $ outputCap f []
 
-infixl 2 <#>
 
--- | An operator version of 'mappend'.
-(<#>) :: Monoid m => m -> m -> m
-(<#>) = mappend
+-- OK, this is the structure that I want:
+class OutputCap f where
+    hasOkPadding :: f -> String -> Bool
+    outputCap :: ([Int] -> String) -> [Int] -> f
 
+instance OutputCap [Char] where
+    hasOkPadding _ = not . strHasPadding 
+    outputCap f xs = f (reverse xs)
+
+instance OutputCap TermOutput where
+    hasOkPadding _ = const True
+    outputCap f xs = fromStr 1 $ f $ reverse xs
+
+instance (Enum p, OutputCap f) => OutputCap (p -> f) where
+    outputCap f xs = \x -> outputCap f (fromEnum x:xs)
+    hasOkPadding _ = hasOkPadding (undefined :: f)
+
+
+{- $outputdoc
+Terminfo contains many string capabilities for special effects.
+For example, the @cuu1@ capability moves the cursor up one line; on ANSI terminals
+this is accomplished by printing the control sequence @\"\\ESC[A\"@.
+However, some older terminals also require \"padding\", or short pauses, after certain commands.
+For example, when @TERM=vt100@ the @cuu1@ capability is @\"\\ESC[A$\<2\>\"@, which instructs terminfo
+to pause for two milliseconds after outputting the control sequence.
+
+The 'TermOutput' monoid abstracts away all padding and control
+sequence output.  Unfortunately, that datatype is difficult to integrate into existing 'String'-based APIs
+such as pretty-printers.  Thus, as a workaround, 'tiGetOutput1' also lets us access the control sequences as 'String's.  The one caveat is that it will not allow you to access padded control sequences as Strings.  For example:
+
+   > > t <- setupTerm "vt100"
+   > > isJust (getCapability t (tiGetOutput1 "cuu1") :: Maybe String)
+   > False
+   > > isJust (getCapability t (tiGetOutput1 "cuu1") :: Maybe TermOutput)
+   > True
+
+'String' capabilities will work with software-based terminal types such as @xterm@ and @linux@.
+However, you should use 'TermOutput' if compatibility with older terminals is important.
+Additionally, the @visualBell@ capability which flashes the screen usually produces its effect with a padding directive, so it will only work with 'TermOutput'.
+
+-}
+
+
+class (Monoid s, OutputCap s) => TermStr s
+
+instance TermStr [Char]
+instance TermStr TermOutput
